#include <pspmoduleinfo.h>
#include <psputils.h>
#include <stdio.h>
#include <string.h>

PSP_MODULE_INFO("benchmark", 0, 0, 71);

static time_t utime_sub(const struct timeval *a, const struct timeval *b);

int main(int argc, char *argv[])
{
	unsigned char __attribute__ ((aligned(16))) buf[256];
	unsigned char *mem;
	int x, y;
	struct timeval before_read, after_read, before_write, after_write;
	time_t readtime, writetime;
	SceKernelUtilsMt19937Context rctx;

	struct memregion_s {
		char *name;
		u32 base;
		int size;
	} regions[] = {
		{ "Scratchpad",			0x00010000, 0x00004000 },
		{ "Scratchpad (uncached)",	0x40010000, 0x00004000 },
		{ "VRAM",			0x04000000, 0x00200000 },
		{ "VRAM (uncached)",		0x44000000, 0x00200000 },
		{ "User memory",		0x08800000, 0x01800000 },
		{ "User memory (uncached)",	0x48800000, 0x01800000 },
//		{ "Internal RAM",		0xBFC00000, 0x00100000 }
	};


	printf("Initializing PRNG...\n");
	sceKernelUtilsMt19937Init(&rctx, 0x42);

	printf("Randomizing...\n");
	for (x = 0; x < sizeof(buf); x++)
		buf[x] = sceKernelUtilsMt19937UInt(&rctx);
	printf("Done\n");


	for (x = 0; x < sizeof(regions) / sizeof(regions[0]); x++) {
		if (strstr(regions[x].name, "Scratchpad"))
			mem = (unsigned char *)regions[x].base;
		else
			mem = (unsigned char *)(regions[x].base + (regions[x].size / 2));
//		printf("Benchmarking '%s' (%p-%p@%p)\n", regions[x].name, (void *)regions[x].base, (void *)(regions[x].base + regions[x].size), mem);

		sceKernelLibcGettimeofday(&before_read, NULL);
		for (y = 0; y < (1024*1024*1024) / sizeof(buf); y++)
			memcpy(buf, mem, sizeof(buf));
		sceKernelDcacheWritebackAll();
		sceKernelLibcGettimeofday(&after_read, NULL);

		sceKernelLibcGettimeofday(&before_write, NULL);
		for (y = 0; y < (1024*1024*1024) / sizeof(buf); y++)
			memcpy(mem, buf, sizeof(buf));
		sceKernelDcacheWritebackAll();
		sceKernelLibcGettimeofday(&after_write, NULL);

		readtime = utime_sub(&after_read, &before_read);
		writetime = utime_sub(&after_write, &before_write);

		printf("%22s/0x%08lx: %8.4f MiB/s read, %8.4f MiB/s write\n", regions[x].name, (unsigned long)mem, 1024000000.0 / (double)readtime, 1024000000.0 / (double)writetime);
	}

	return 0;
}

static time_t utime_sub(const struct timeval *a, const struct timeval *b)
{
	return ((a->tv_sec - b->tv_sec) * 1000000) + (a->tv_usec - b->tv_usec);
}

/*
            Scratchpad/0x00010000:  63.7137 MiB/s read,  59.6770 MiB/s write
 Scratchpad (uncached)/0x40010000:  26.0393 MiB/s read,  45.5564 MiB/s write
                  VRAM/0x04100000:  60.9029 MiB/s read,  56.9173 MiB/s write
       VRAM (uncached)/0x44100000:  12.1767 MiB/s read,  45.3064 MiB/s write
           User memory/0x09400000:  47.7323 MiB/s read,  47.6334 MiB/s write
User memory (uncached)/0x49400000:   9.4286 MiB/s read,  58.7772 MiB/s write

*/
